Attribute VB_Name = "mcrMain"
Option Explicit

Public Type ConfigType
    ConfigString As String
    Language As String
    MainLeft As Integer
    MainTop As Integer
    ImageOpen As String
    ImageSave As String
    Author As String
End Type

Public Type LanguageType
    FileLength As String
    ImageHeight As String
    Success As String
    PalettedWarning As String
    ImageSize As String
End Type

'header: 14 bytes
Private Type HeaderType
     FileType As String * 2         'file type, should be 4D42h or BM
     FileSize As Long               'compress amount, should be 0
     Reserved1 As Integer           'always 0
     Reserved2 As Integer           'always 0
     ImageDataLocation As Long      'location of image data in bytes
End Type


'image info: 40 bytes
Private Type InfoType
     HeaderSize As Long             'size of header
     imageWidth As Long             'width of image
     ImageHeight As Long            'height of image
     NumPlanes As Integer           'flat images are 1
     NumPixelBits As Integer        'number of bits per pixel: 1, 4, 8, or 24
     CompressionType As Long        'bitmaps should not be compressed, so this should be 0
     ImageSize As Long              'in bytes
     X_Pixels_per_Meter As Long     'preferred X-axis resolution in pixels per meter
     Y_Pixels_per_Meter As Long     'preferred Y-axis resolution in pixels per meter
     NumColors As Long              'number of colors that are used
     PrimaryColor As Long           'which color is dominant, 0 is all colors
End Type

'color palette: 4 bytes * 256 = 1024
Private Type PaletteType
     NumBlue As Byte
     NumGreen As Byte
     NumRed As Byte
     NumOther As Byte
End Type

'LDRAW part file type
Private Type LdrawType
    numPoints As Integer
    colorNum As Long
    X(1 To 4) As Double
    Y(1 To 4) As Double
    Z(1 To 4) As Double
End Type

Private BMPHeader As HeaderType
Public BMPInfo As InfoType
Private BMPPalette As PaletteType
Private ColorPalette() As PaletteType
Private ColorArray(0 To 15) As PaletteType
Private RawImageData() As Byte
Public LdrawImageData() As LdrawType
Public strAuthor As String

Public LanguageData As LanguageType
Public ConfigData As ConfigType
Public LOGGER As TextStream
Public LoggerFile As FileSystemObject

Public Const PI = 3.14159265

Public Sub LoadBitmap(BitmapLocation As String, ImageHeight As Double, imageWidth As Double)
    Dim i As Long
    Dim j As Long
    Dim strColor As String
    Dim bytColor As Byte
    
    On Error GoTo errHandler
    
    Call WriteToLog("LOAD BITMAP: Starting")
    
    'read header data
    Open BitmapLocation For Binary As #1
    Call WriteToLog("LOAD BITMAP: About to read header")
    Get #1, , BMPHeader
    Call WriteToLog("LOAD BITMAP: About to read info")
    Get #1, , BMPInfo
    
    'read palette data
    Call WriteToLog("LOAD BITMAP: About to read color palette")
    For i = 0 To (BMPHeader.ImageDataLocation - 54) / Len(BMPPalette) - 1
        ReDim Preserve ColorPalette(i)
        Get #1, , ColorPalette(i)
        'Call WriteToLog("LOAD BITMAP: " & i & ": " & ColorPalette(i).NumRed & " " & ColorPalette(i).NumGreen & " " & ColorPalette(i).NumBlue)
    Next i
    If i = 0 Then
        ReDim ColorPalette(255)
        Call WriteToLog("LOAD BITMAP: color palette empty")
    End If
    
    If i = 0 Or i = 2 Then
        'MsgBox "Warning, the image that you have loaded may not be a paletted bitmap"
        MsgBox LanguageData.PalettedWarning
    End If
    
    'read image data
    Call WriteToLog("LOAD BITMAP: About to read raw image data")
    ReDim RawImageData(BMPInfo.imageWidth, BMPInfo.ImageHeight)
    For i = 1 To BMPInfo.ImageHeight
        For j = 1 To BMPInfo.imageWidth
            If Not EOF(1) Then
                Get #1, , RawImageData(j, i)
            End If
        Next j
        For j = 1 To (BMPInfo.ImageSize / BMPInfo.ImageHeight) - BMPInfo.imageWidth 'bitmap format may have padding on the width, so the width can be divisble by 4
            Get #1, , bytColor
        Next j
    Next i
    Close #1
    
    imageWidth = BMPInfo.imageWidth
    ImageHeight = BMPInfo.ImageHeight
    If ImageHeight > 1024 Or imageWidth > 1024 Then
        'MsgBox "This image could be too large.  Please resize it smaller than 1024 pixels in height / width"
        MsgBox LanguageData.ImageSize
        Call WriteToLog("LOAD BITMAP: large image warning: Height: " & ImageHeight & " Width: " & imageWidth)
    End If
    Call WriteToLog("LOAD BITMAP: Done")
    Exit Sub
    
errHandler:
    MsgBox "LoadBitmap has caused an error " & Err.Description
    Call WriteToLog("LOAD BITMAP: ERROR LOADING BITMAP " & Err.Description)
    Close #1
End Sub

Public Sub CreateBitmap(SaveLocation As String, LDRAWColors As Boolean, IgnoreWhite As Boolean, CloseToWhite As Long, IgnoreColor As Boolean, CloseToColor As Long, userRed As Integer, userGreen As Integer, userBlue As Integer)
    Dim i As Integer
    Dim j As Integer
    Dim temp As Integer
    Dim tempColor As PaletteType
    Dim NewHeader As HeaderType
    Dim NewInfo As InfoType
    Dim bytColor As Byte
    Dim lngColor As Long
    
    On Error GoTo ErrorHandler
    
    Call WriteToLog("CREATE BITMAP: Starting")
    
    Call WriteToLog("CREATE BITMAP: About to open save location for writing")
    Open SaveLocation For Binary As #2
    
    'assign values
    NewHeader.FileType = BMPHeader.FileType
    NewHeader.FileSize = BMPHeader.FileSize
    NewHeader.Reserved1 = BMPHeader.Reserved1
    NewHeader.Reserved2 = BMPHeader.Reserved2
    NewHeader.ImageDataLocation = BMPHeader.ImageDataLocation
    
    NewInfo.HeaderSize = BMPInfo.HeaderSize
    NewInfo.imageWidth = BMPInfo.imageWidth
    NewInfo.ImageHeight = BMPInfo.ImageHeight
    NewInfo.NumPlanes = BMPInfo.NumPlanes
    NewInfo.NumPixelBits = BMPInfo.NumPixelBits
    NewInfo.CompressionType = BMPInfo.CompressionType
    NewInfo.ImageSize = BMPInfo.ImageSize
    NewInfo.X_Pixels_per_Meter = BMPInfo.X_Pixels_per_Meter
    NewInfo.Y_Pixels_per_Meter = BMPInfo.Y_Pixels_per_Meter
    NewInfo.NumColors = BMPInfo.NumColors
    NewInfo.PrimaryColor = BMPInfo.PrimaryColor
    
    'try to find gray in the colorpalette
    If IgnoreWhite Or IgnoreColor Then
        temp = FindGray
        'if gray is not there, then we need to add it
        If temp = 0 Then
            If UBound(ColorPalette) < 255 Then
                temp = UBound(ColorPalette) + 1
                ReDim Preserve ColorPalette(temp)
                NewHeader.FileSize = NewHeader.FileSize + 4
                NewHeader.ImageDataLocation = NewHeader.ImageDataLocation + 4
                NewInfo.NumColors = NewInfo.NumColors + 1
            Else    'we have to overwrite one of the palette entries, but this is just for the preview image, and does not effect the LDRAW file
                temp = 255
            End If
            ColorPalette(temp).NumRed = 212
            ColorPalette(temp).NumGreen = 208
            ColorPalette(temp).NumBlue = 200
            ColorPalette(temp).NumOther = 0
        End If
    End If
    
    'save file header
    Call WriteToLog("CREATE BITMAP: About to write header and info")
    Put #2, , NewHeader
    Put #2, , NewInfo
        
    'set color palette and write to file
    If LDRAWColors Then
        Call SetColorPalette
    End If
    Call WriteToLog("CREATE BITMAP: About to write color palette")
    For i = 0 To (NewHeader.ImageDataLocation - 54) / Len(BMPPalette) - 1
        If i < 16 And LDRAWColors Then
            Put #2, , ColorArray(i)
        Else
            Put #2, , ColorPalette(i)
        End If
    Next i
    
    'write the raw image data
    Call WriteToLog("CREATE BITMAP: About to put raw image data")
    For i = 1 To BMPInfo.ImageHeight
        For j = 1 To BMPInfo.imageWidth
            lngColor = ReColor(RawImageData(j, i), LDRAWColors)
            'if the user selected to make white or other color transparent, we need to write gray instead
            If (IgnoreWhite And IsColorWhite(RawImageData(j, i), CloseToWhite)) Or (IgnoreColor And IsThisColorCloseEnough(RawImageData(j, i), CloseToColor, userRed, userGreen, userBlue)) Then
                bytColor = temp
                Put #2, , bytColor
            ElseIf LDRAWColors Then
                Put #2, , CByte(lngColor)
            Else
                Put #2, , RawImageData(j, i)
            End If
        Next j
        For j = 1 To (NewInfo.ImageSize / NewInfo.ImageHeight) - NewInfo.imageWidth 'bitmap format requires the width be divisible by 4, so we may need to pad it
            bytColor = 0
            Put #2, , bytColor
        Next j
    Next i
    Call WriteToLog("CREATE BITMAP: About to close file")
    Close #2
    Call WriteToLog("CREATE BITMAP: Done")
    Exit Sub
    
ErrorHandler:
    MsgBox "CreateBitmap has caused an error " & Err.Description
    Call WriteToLog("CREATE BITMAP: ERROR CREATING TEMP BITMAP " & Err.Description)
    Close #2
End Sub

Public Sub ConvertToLdraw(RequestedHeight As Double, LDRAWColors As Boolean, IgnoreWhite As Boolean, CloseToWhite As Long, IgnoreColor As Boolean, CloseToColor As Long, userRed As Integer, userGreen As Integer, userBlue As Integer)
    Dim i As Integer
    Dim j As Integer
    Dim k As Integer
    Dim m As Integer
    Dim n As Integer
    Dim ndx As Integer
    Dim row As Integer
    Dim tempImageData() As Long
    Dim tempObjectData() As Integer
    Dim tempColor As Long
    Dim lngColor As Long
    Dim X(1 To 4) As Double
    Dim Y(1 To 4) As Double
    Dim maxX(1 To 4) As Double
    Dim maxY(1 To 4) As Double
    Dim maxArea As Double
    Dim tempDouble As Double
    Dim minH As Integer
    Dim maxH As Integer
    Dim minV As Integer
    Dim maxV As Integer
    Dim lastrow As Integer
    
    On Error GoTo errHandler
    
    'bitmaps are written upside down, so we are going to flip it and apply the user requested color changes
    Call SetColorPalette
    ReDim tempImageData(BMPInfo.imageWidth, BMPInfo.ImageHeight)
    ndx = 0
    For i = BMPInfo.ImageHeight To 1 Step -1
        ndx = ndx + 1
        For j = 1 To BMPInfo.imageWidth
            lngColor = ReColor(RawImageData(j, i), LDRAWColors)
            If ((IgnoreWhite And IsColorWhite(RawImageData(j, i), CloseToWhite)) = False) And ((IgnoreColor And IsThisColorCloseEnough(RawImageData(j, i), CloseToColor, userRed, userGreen, userBlue)) = False) Then
            'If ((IgnoreWhite And (lngColor > CloseToWhite)) = False) And ((IgnoreColor And (lngColor < CloseToColor)) = False) Then
                tempImageData(j, ndx) = lngColor
            Else
                tempImageData(j, ndx) = 999
            End If
        Next j
    Next i
    
    ndx = 0
    For m = 1 To BMPInfo.ImageHeight
        For n = 1 To BMPInfo.imageWidth
            If Not tempImageData(n, m) = 999 Then
                tempColor = tempImageData(n, m)
                'find the edges of regions with the same color, looking row by row
                
                ReDim tempObjectData(BMPInfo.ImageHeight, 4)
                lastrow = 0
                For i = 1 To BMPInfo.ImageHeight
                    If (lastrow = 0) Or (i - lastrow) < 2 Then 'if there is a whole row without our color, stop looking
                        'look for the first occurance of our color to find the first edge of the region
                        For j = 1 To BMPInfo.imageWidth
                            If tempImageData(j, i) = tempColor Then
                                tempObjectData(i, 1) = j
                                Exit For
                            End If
                        Next j
                        'look for the color change to find the second edge of the region
                        For k = j To BMPInfo.imageWidth
                            If tempImageData(k, i) = tempColor Then 'if the color is the same, mark it, if not, move to the next row
                                tempImageData(k, i) = 999
                                lastrow = i
                            Else
                                Exit For
                            End If
                        Next k
                        tempObjectData(i, 2) = k - 1
                    Else
                        Exit For
                    End If
                Next i
                
                'find the max and min values surounding the region
                minV = 32000
                maxV = 0
                For i = 1 To BMPInfo.ImageHeight
                    If tempObjectData(i, 1) = 0 Then
                        If (Not minV = 32000) And (maxV = 0) Then
                            maxV = i - 1
                            Exit For
                        End If
                    ElseIf minV = 32000 Then
                        minV = i
                    End If
                Next i
                If maxV = 0 Then
                    maxV = BMPInfo.ImageHeight
                End If
                minH = 32000
                maxH = 0
                For i = minV To maxV
                    If tempObjectData(i, 1) < minH Then
                        minH = tempObjectData(i, 1)
                    End If
                    If tempObjectData(i, 2) > maxH Then
                        maxH = tempObjectData(i, 2)
                    End If
                Next i
                
                If ((maxV - minV) < 0) Or ((maxH - minH) < 0) Then
                    GoTo SkipThis
                End If
                
                'now that we have a complete region, we need to find the largest quads to fit in it
                For i = minV To maxV
                    If Not tempObjectData(i, 1) = 0 Then    'if we have not looked at it yet
                        X(1) = tempObjectData(i, 1)
                        Y(1) = i
                        X(2) = tempObjectData(i, 2) + 0.0000001
                        Y(2) = i
                        'look for the largest area that we can get for the third and fourth point
                        maxArea = -1
                        For j = i To maxV
                            If Not tempObjectData(j, 1) = 0 Then
                                X(3) = tempObjectData(j, 1)
                                Y(3) = j + 0.0000001
                                For k = i To maxV
                                    X(4) = tempObjectData(k, 2) + 0.0000001
                                    Y(4) = k + 0.0000001
                                    'if any of the angles of the polygon are too sharp (like a line), don't use it
                                    'If TestAngles(X, Y) Then
                                        If AreaHasSameColor(tempObjectData, minV, maxV, minH, maxH, X, Y) Then
                                            tempDouble = 0.5 * Abs(CrossProduct_2D(X, Y))
                                            If Y(3) = Y(4) Then
                                                tempDouble = tempDouble * 1.2
                                            End If
                                            If X(2) = X(4) Then
                                                tempDouble = tempDouble * 1.2
                                            End If
                                            'tempDouble = 0.5 * Abs(CrossProductLength())
                                            If tempDouble > maxArea Then
                                                maxArea = tempDouble
                                                maxX(3) = X(3)
                                                maxY(3) = Y(3)
                                                maxX(4) = X(4)
                                                maxY(4) = Y(4)
                                            End If
                                        End If
                                    'End If
                                Next k
                            End If
                        Next j
                        X(2) = X(2) - 0.0000001
                        X(3) = maxX(3)
                        Y(3) = maxY(3) - 0.0000001
                        X(4) = maxX(4) - 0.0000001
                        Y(4) = maxY(4) - 0.0000001
                        If Not maxArea = 0 Then
                            ndx = ndx + 1
                            ReDim Preserve LdrawImageData(ndx)
                            LdrawImageData(ndx).numPoints = 4
                            LdrawImageData(ndx).colorNum = tempColor
                            If (Not CInt(X(1)) = CInt(X(2))) Or (maxArea < 2) Then
                                LdrawImageData(ndx).X(1) = X(1)
                                LdrawImageData(ndx).Y(1) = Y(1)
                                LdrawImageData(ndx).X(2) = X(2) + 1
                                LdrawImageData(ndx).Y(2) = Y(2)
                                LdrawImageData(ndx).X(3) = X(3)
                                LdrawImageData(ndx).Y(3) = Y(3) + 1
                                LdrawImageData(ndx).X(4) = X(4) + 1
                                LdrawImageData(ndx).Y(4) = Y(4) + 1
                            Else
                                LdrawImageData(ndx).X(1) = X(1)
                                LdrawImageData(ndx).Y(1) = Y(1)
                                LdrawImageData(ndx).X(2) = X(2)
                                LdrawImageData(ndx).Y(2) = Y(2)
                                LdrawImageData(ndx).X(3) = X(3)
                                LdrawImageData(ndx).Y(3) = Y(3) + 1
                                LdrawImageData(ndx).X(4) = X(4) + 1
                                LdrawImageData(ndx).Y(4) = Y(4) + 0.5
                            End If
                        End If
                        'now that have chosen the largest quad or triangle, we need to mark the area inside as being recorded
                        Call MarkArea(tempObjectData, minV, maxV, minH, maxH, X, Y)
                    End If
                Next i
            End If
SkipThis:
        Next n
    Next m
    
    Exit Sub
    
errHandler:
    MsgBox "ConvertToLdraw has caused an error " & Err.Description
    Call WriteToLog("CONVERTTOLDRAW: ERROR CONVERTING BITMAP " & Err.Description)
End Sub

Private Function CrossProduct_2D(X() As Double, Y() As Double) As Double
    CrossProduct_2D = ((X(3) - X(2)) * (Y(4) - Y(1))) - ((Y(3) - Y(2)) * (X(4) - X(1)))
End Function

Private Function MarkArea(ImageData() As Integer, minV As Integer, maxV As Integer, minH As Integer, maxH As Integer, X() As Double, Y() As Double)
    Dim i As Integer
    Dim j As Integer
    Dim firstIn As Integer
    Dim firstOut As Integer
    
    For i = minV To maxV
        firstIn = 32000
        firstOut = 0
        For j = minH To maxH
            If PointInPolygon2(j, i, X, Y) Then
                If firstIn = 32000 Then
                    firstIn = j
                End If
            ElseIf (Not firstIn = 32000) And firstOut = 0 Then
                firstOut = j
            End If
        Next j
        If firstOut = 0 Then
            firstOut = maxH
        End If
        If firstIn = 32000 Then
            'do nothing
        ElseIf firstIn <= ImageData(i, 1) And firstOut >= ImageData(i, 2) Then
            ImageData(i, 1) = 0
        ElseIf firstIn <= ImageData(i, 1) Then
            ImageData(i, 1) = firstOut
        ElseIf firstOut >= ImageData(i, 2) Then
            ImageData(i, 2) = firstIn
        ElseIf firstIn > ImageData(i, 1) And firstOut < ImageData(i, 2) Then
            ImageData(i, 4) = ImageData(i, 2)
            ImageData(i, 2) = firstIn
            ImageData(i, 3) = firstOut
        End If
    Next i
End Function

Private Function AreaHasSameColor(ImageData() As Integer, minV As Integer, maxV As Integer, minH As Integer, maxH As Integer, X() As Double, Y() As Double) As Boolean
    Dim i As Integer
    Dim j As Integer
    
    AreaHasSameColor = False
    For i = minV To maxV
        If Not ImageData(i, 1) = 0 Then
            For j = minH To maxH
                'if this pixel is in the bounded quad, then see if it is in our region of color
                If PointInPolygon1(j, i, X, Y) Then
                    If ((j < ImageData(i, 1)) Or (j > ImageData(i, 2))) Then ' And (Not ImageData(i, 1) = ImageData(i, 2)) Then
                        Exit Function
                    End If
                End If
            Next j
        End If
    Next i
    AreaHasSameColor = True
    Exit Function
    
errHandler:
    MsgBox "AreaHasSameColor has caused an error " & Err.Description
    Call WriteToLog("AREAHASSAMECOLOR: ERROR CONVERTING BITMAP " & Err.Description)
End Function

Private Function TestAngles(X() As Double, Y() As Double) As Boolean
    Dim px(1 To 4) As Single
    Dim py(1 To 4) As Single
    TestAngles = False
    
    px(1) = X(1) - 0.00000001
    py(1) = Y(1) - 0.00000001
    px(2) = X(2) + 0.00000001
    py(2) = Y(2) - 0.00000001
    px(3) = X(4) + 0.00000001
    py(3) = Y(4) + 0.00000001
    px(4) = X(3) - 0.00000001
    py(4) = Y(3) + 0.00000001
    
    If GetAngle(px(1), py(1), px(2), py(2), px(3), py(3)) < 5 Then
        Exit Function
    End If
    
    If GetAngle(px(2), py(2), px(3), py(3), px(4), py(4)) < 5 Then
        Exit Function
    End If
    
    If GetAngle(px(3), py(3), px(4), py(4), px(1), py(1)) < 5 Then
        Exit Function
    End If
    
    If GetAngle(px(4), py(4), px(1), py(1), px(2), py(2)) < 5 Then
        Exit Function
    End If
    TestAngles = True
End Function

Public Function PointInPolygon1(PointX As Integer, PointY As Integer, X() As Double, Y() As Double) As Boolean
    Dim i As Integer
    Dim testX As Single
    Dim testY As Single
    Dim px(1 To 4) As Single
    Dim py(1 To 4) As Single
    Dim total_angle As Single

    testX = PointX
    testY = PointY
    
    px(1) = X(1) - 0.5
    py(1) = Y(1) - 0.5
    px(2) = X(2) + 0.5
    py(2) = Y(2) - 0.5
    px(3) = X(4) + 0.5
    py(3) = Y(4) + 0.5
    px(4) = X(3) - 0.5
    py(4) = Y(3) + 0.5

    ' Get the angle between the point and the first and last vertices
    total_angle = GetAngle(px(4), py(4), testX, testY, px(1), py(1))

    ' Add the angles from the point to each other pair of vertices.
    For i = 1 To 3
        total_angle = total_angle + GetAngle(px(i), py(i), testX, testY, px(i + 1), py(i + 1))
    Next i

    ' If the point is in the polygon the total angle should be 2 * PI or -2 * PI, and close to zero if the point is outside the polygon
    PointInPolygon1 = (Abs(total_angle) > PI)
End Function

Public Function PointInPolygon2(PointX As Integer, PointY As Integer, X() As Double, Y() As Double) As Boolean
    Dim i As Integer
    Dim testX As Single
    Dim testY As Single
    Dim px(1 To 4) As Single
    Dim py(1 To 4) As Single
    Dim total_angle As Single

    testX = PointX
    testY = PointY
    
    px(1) = X(1) - 0.1
    py(1) = Y(1) - 0.1
    px(2) = X(2) + 0.1
    py(2) = Y(2) - 0.1
    px(3) = X(4) + 0.1
    py(3) = Y(4) + 0.1
    px(4) = X(3) - 0.1
    py(4) = Y(3) + 0.1

    ' Get the angle between the point and the first and last vertices
    total_angle = GetAngle(px(4), py(4), testX, testY, px(1), py(1))

    ' Add the angles from the point to each other pair of vertices.
    For i = 1 To 3
        total_angle = total_angle + GetAngle(px(i), py(i), testX, testY, px(i + 1), py(i + 1))
    Next i

    ' If the point is in the polygon the total angle should be 2 * PI or -2 * PI, and close to zero if the point is outside the polygon
    PointInPolygon2 = (Abs(total_angle) > PI)
End Function

Public Function GetAngle(Ax As Single, Ay As Single, Bx As Single, By As Single, Cx As Single, Cy As Single) As Single
    Dim dot_product As Single
    Dim cross_product As Single

    ' Get the dot product and cross product.
    dot_product = DotProduct(Ax, Ay, Bx, By, Cx, Cy)
    cross_product = CrossProductLength(Ax, Ay, Bx, By, Cx, Cy)

    ' Calculate the angle.
    GetAngle = ATan2(cross_product, dot_product)
End Function

Public Function ATan2(opp As Single, adj As Single) As Single
    Dim angle As Single

    ' Get the basic angle.
    If Abs(adj) < 0.0001 Then
        angle = PI / 2
    Else
        angle = Abs(Atn(opp / adj))
    End If

    ' See if we are in quadrant 2 or 3.
    If adj < 0 Then
        ' angle > PI/2 or angle < -PI/2.
        angle = PI - angle
    End If

    ' See if we are in quadrant 3 or 4.
    If opp < 0 Then
        angle = -angle
    End If

    ' Return the result.
    ATan2 = angle
End Function

Private Function DotProduct(Ax As Single, Ay As Single, Bx As Single, By As Single, Cx As Single, Cy As Single) As Single
    Dim BAx As Single
    Dim BAy As Single
    Dim BCx As Single
    Dim BCy As Single

    ' Get the vectors' coordinates.
    BAx = Ax - Bx
    BAy = Ay - By
    BCx = Cx - Bx
    BCy = Cy - By

    ' Calculate the dot product.
    DotProduct = BAx * BCx + BAy * BCy
End Function

Public Function CrossProductLength(Ax As Single, Ay As Single, Bx As Single, By As Single, Cx As Single, Cy As Single) As Single
    Dim BAx As Single
    Dim BAy As Single
    Dim BCx As Single
    Dim BCy As Single

    ' Get the vectors' coordinates.
    BAx = Ax - Bx
    BAy = Ay - By
    BCx = Cx - Bx
    BCy = Cy - By

    ' Calculate the Z coordinate of the cross product.
    CrossProductLength = BAx * BCy - BAy * BCx
End Function

Public Sub WritePartFile(SaveLocation As String, RequestedHeight As Long, MetaData() As String)
    Dim i As Long
    Dim writer As TextStream
    Dim OutFile As FileSystemObject
    Dim BitSize As Double
    BitSize = RequestedHeight / BMPInfo.ImageHeight

    On Error GoTo errHandler
    
    Call WriteToLog("WRITE PART FILE: Starting")
    
    'create output part file
    Call WriteToLog("WRITE PART FILE: About to create output file")
    Set OutFile = New FileSystemObject
    Set writer = OutFile.OpenTextFile(SaveLocation, ForWriting, True)
    
    'write meta data to file
    Call WriteToLog("WRITE PART FILE: About to write to file")
    writer.WriteLine ("0 " & MetaData(0) & ".dat")
    writer.WriteLine ("0 Name: " & MetaData(1))
    writer.WriteLine ("0 Author: " & MetaData(2))
    writer.WriteLine ("0 " & MetaData(3))

    'write raw image data to file - resize according to requested size
    Call WriteToLog("WRITE PART FILE: About to write to raw image data")
    For i = 1 To UBound(LdrawImageData)
        writer.WriteLine (LdrawImageData(i).numPoints & " " & LdrawImageData(i).colorNum & " " & LdrawImageData(i).X(1) * BitSize & " " & LdrawImageData(i).Y(1) * BitSize & " 0 " & LdrawImageData(i).X(2) * BitSize & " " & LdrawImageData(i).Y(2) * BitSize & " 0 " & LdrawImageData(i).X(3) * BitSize & " " & LdrawImageData(i).Y(3) * BitSize & " 0 " & LdrawImageData(i).X(4) * BitSize & " " & LdrawImageData(i).Y(4) * BitSize & " 0")
    Next i
    Call WriteToLog("WRITE PART FILE: About to close file")
    writer.Close
    Call WriteToLog("WRITE PART FILE: Done")
    Exit Sub
    
errHandler:
    MsgBox "WritePartFile has caused an error " & Err.Description
    Call WriteToLog("WRITE PART FILE: ERROR CONVERTING BITMAP " & Err.Description)
    writer.Close
End Sub

Public Sub ConvertBitmap(SaveLocation As String, LDRAWColors As Boolean, IgnoreWhite As Boolean, CloseToWhite As Long, IgnoreColor As Boolean, CloseToColor As Long, userRed As Integer, userGreen As Integer, userBlue As Integer)
    Dim i As Long
    Dim j As Long
    Dim ndx As Long
    Dim rowndx As Double
    Dim colndx As Double
    Dim X(1 To 4) As Double
    Dim Y(1 To 4) As Double
    Dim lngColor As Long
    Dim writer As TextStream
    Dim OutFile As FileSystemObject
    Dim bytColor As Byte
    
    On Error GoTo errHandler
    
    Call WriteToLog("CONVERT FILE: Starting")
    
    'write raw image data to file
    Call WriteToLog("CONVERT FILE: About to write to raw image data")
    colndx = 1
    For i = BMPInfo.ImageHeight To 1 Step -1    'bitmaps are upside down, so we need to write the part file right side up
        X(1) = 1
        X(4) = 1
        X(2) = 2
        X(3) = 2
        Y(1) = colndx
        Y(2) = colndx
        Y(3) = colndx + 1
        Y(4) = colndx + 1
        rowndx = 1
        For j = 1 To BMPInfo.imageWidth
            rowndx = rowndx + 1
            lngColor = ReColor(RawImageData(j, i), LDRAWColors)
            If Not j = BMPInfo.imageWidth Then
                'if this pixel is the same color as the next pixel, combine them into a line instead of writing each pixel
                If Not RawImageData(j + 1, i) = RawImageData(j, i) Then
                    'If ((IgnoreWhite And (lngColor > CloseToWhite)) = False) And ((IgnoreColor And (lngColor < CloseToColor)) = False) Then
                    If ((IgnoreWhite And IsColorWhite(RawImageData(j, i), CloseToWhite)) = False) And ((IgnoreColor And IsThisColorCloseEnough(RawImageData(j, i), CloseToColor, userRed, userGreen, userBlue)) = False) Then
                        ndx = ndx + 1
                        ReDim Preserve LdrawImageData(ndx)
                        LdrawImageData(ndx).numPoints = 4
                        LdrawImageData(ndx).colorNum = lngColor
                        LdrawImageData(ndx).X(1) = X(1)
                        LdrawImageData(ndx).Y(1) = Y(1)
                        LdrawImageData(ndx).X(2) = X(2)
                        LdrawImageData(ndx).Y(2) = Y(2)
                        LdrawImageData(ndx).X(3) = X(3)
                        LdrawImageData(ndx).Y(3) = Y(3)
                        LdrawImageData(ndx).X(4) = X(4)
                        LdrawImageData(ndx).Y(4) = Y(4)
                    End If
                    X(1) = rowndx
                    X(4) = rowndx
                    X(2) = rowndx + 1
                    X(3) = rowndx + 1
                Else
                    X(2) = X(2) + 1
                    X(3) = X(2)
                End If
            Else
                'If ((IgnoreWhite And (lngColor > CloseToWhite)) = False) And ((IgnoreColor And (lngColor < CloseToColor)) = False) Then
                If ((IgnoreWhite And IsColorWhite(RawImageData(j, i), CloseToWhite)) = False) And ((IgnoreColor And IsThisColorCloseEnough(RawImageData(j, i), CloseToColor, userRed, userGreen, userBlue)) = False) Then
                    ndx = ndx + 1
                    ReDim Preserve LdrawImageData(ndx)
                    LdrawImageData(ndx).numPoints = 4
                    LdrawImageData(ndx).colorNum = lngColor
                    LdrawImageData(ndx).X(1) = X(1)
                    LdrawImageData(ndx).Y(1) = Y(1)
                    LdrawImageData(ndx).X(2) = X(2)
                    LdrawImageData(ndx).Y(2) = Y(2)
                    LdrawImageData(ndx).X(3) = X(3)
                    LdrawImageData(ndx).Y(3) = Y(3)
                    LdrawImageData(ndx).X(4) = X(4)
                    LdrawImageData(ndx).Y(4) = Y(4)
                End If
            End If
        Next j
        colndx = colndx + 1
    Next i
    Call WriteToLog("CONVERT FILE: Done")
    Exit Sub
    
errHandler:
    MsgBox "ConvertBitmap has caused an error " & Err.Description
    Call WriteToLog("CONVERT FILE: ERROR CONVERTING BITMAP " & Err.Description)
    writer.Close
End Sub

Private Function ReColor(mycolor As Byte, LDRAWColors As Boolean) As Long
    Dim r As String
    Dim g As String
    Dim b As String
    Dim NewColor As String
    Dim i As Integer
    Dim intR1 As Integer
    Dim intR2 As Integer
    Dim intG1 As Integer
    Dim intG2 As Integer
    Dim intB1 As Integer
    Dim intB2 As Integer
    Dim tempColor As Integer
    Dim minColorDiff As Integer
    Dim tempColorDiff As Integer
    
    On Error GoTo errHandler
    
    r = Hex(ColorPalette(mycolor).NumRed)
    g = Hex(ColorPalette(mycolor).NumGreen)
    b = Hex(ColorPalette(mycolor).NumBlue)
    
    If Len(r) = 1 Then
        r = "0" & r
    End If
    If Len(g) = 1 Then
        g = "0" & g
    End If
    If Len(b) = 1 Then
        b = "0" & b
    End If
    
    NewColor = r & g & b
    
    ReColor = (2 * Val("&HFFFFFF&")) + (Val("&H" & NewColor & "&")) + 2
    
    'if the user requests to use LDRAW primary colors then manipulate the output color
    'find the closest primary color to the image color
    If LDRAWColors Then
        minColorDiff = 32000
        tempColor = 0
        For i = 0 To 15
            intR1 = ColorPalette(mycolor).NumRed
            intR2 = ColorArray(i).NumRed
            intG1 = ColorPalette(mycolor).NumGreen
            intG2 = ColorArray(i).NumGreen
            intB1 = ColorPalette(mycolor).NumBlue
            intB2 = ColorArray(i).NumBlue
            tempColorDiff = Abs(intR1 - intR2) + Abs(intG1 - intG2) + Abs(intB1 - intB2)
            If tempColorDiff < minColorDiff Then
                minColorDiff = tempColorDiff
                tempColor = i
            End If
        Next i
        ReColor = tempColor
    End If
    Exit Function
    
errHandler:
    MsgBox "ReColor has caused an error " & Err.Description
    Call WriteToLog("RECOLOR: ERROR IN CONVERTING COLOR " & Err.Description)
End Function

Private Sub SetColorPalette()
    ColorArray(0).NumRed = 0
    ColorArray(0).NumGreen = 16
    ColorArray(0).NumBlue = 14
    ColorArray(1).NumRed = 0
    ColorArray(1).NumGreen = 51
    ColorArray(1).NumBlue = 187
    ColorArray(2).NumRed = 0
    ColorArray(2).NumGreen = 119
    ColorArray(2).NumBlue = 51
    ColorArray(3).NumRed = 0
    ColorArray(3).NumGreen = 170
    ColorArray(3).NumBlue = 170
    ColorArray(4).NumRed = 204
    ColorArray(4).NumGreen = 0
    ColorArray(4).NumBlue = 0
    ColorArray(5).NumRed = 255
    ColorArray(5).NumGreen = 51
    ColorArray(5).NumBlue = 153
    ColorArray(6).NumRed = 102
    ColorArray(6).NumGreen = 51
    ColorArray(6).NumBlue = 0
    ColorArray(7).NumRed = 170
    ColorArray(7).NumGreen = 170
    ColorArray(7).NumBlue = 170
    ColorArray(8).NumRed = 102
    ColorArray(8).NumGreen = 102
    ColorArray(8).NumBlue = 85
    ColorArray(9).NumRed = 0
    ColorArray(9).NumGreen = 136
    ColorArray(9).NumBlue = 255
    ColorArray(10).NumRed = 51
    ColorArray(10).NumGreen = 255
    ColorArray(10).NumBlue = 102
    ColorArray(11).NumRed = 85
    ColorArray(11).NumGreen = 170
    ColorArray(11).NumBlue = 255
    ColorArray(12).NumRed = 255
    ColorArray(12).NumGreen = 85
    ColorArray(12).NumBlue = 85
    ColorArray(13).NumRed = 255
    ColorArray(13).NumGreen = 187
    ColorArray(13).NumBlue = 204
    ColorArray(14).NumRed = 255
    ColorArray(14).NumGreen = 238
    ColorArray(14).NumBlue = 0
    ColorArray(15).NumRed = 255
    ColorArray(15).NumGreen = 255
    ColorArray(15).NumBlue = 255
End Sub

Public Function IsColorWhite(i As Byte, minColorDiff As Long) As Boolean
    Dim intR1 As Integer
    Dim intG1 As Integer
    Dim intB1 As Integer
    Dim tempColorDiff As Integer
    
    IsColorWhite = False
    
    intR1 = ColorPalette(i).NumRed
    intG1 = ColorPalette(i).NumGreen
    intB1 = ColorPalette(i).NumBlue
    tempColorDiff = Abs(intR1 - 255) + Abs(intG1 - 255) + Abs(intB1 - 255)
    If tempColorDiff < minColorDiff Then
        IsColorWhite = True
    End If
End Function

Public Function IsThisColorCloseEnough(i As Byte, minColorDiff As Long, userRed As Integer, userGreen As Integer, userBlue As Integer) As Boolean
    Dim intR1 As Integer
    Dim intG1 As Integer
    Dim intB1 As Integer
    Dim tempColorDiff As Integer
    
    IsThisColorCloseEnough = False
    
    intR1 = ColorPalette(i).NumRed
    intG1 = ColorPalette(i).NumGreen
    intB1 = ColorPalette(i).NumBlue
    tempColorDiff = Abs(intR1 - userRed) + Abs(intG1 - userGreen) + Abs(intB1 - userBlue)
    If tempColorDiff < minColorDiff Then
        IsThisColorCloseEnough = True
    End If
End Function

Public Function FindGray()
    Dim i As Integer
    Dim intR1 As Integer
    Dim intG1 As Integer
    Dim intB1 As Integer
    Dim minColorDiff As Integer
    Dim tempColorDiff As Integer
    
    'find the closest color in the palette to gray
    minColorDiff = 32000
    For i = 0 To UBound(ColorPalette)
        intR1 = ColorPalette(i).NumRed
        intG1 = ColorPalette(i).NumGreen
        intB1 = ColorPalette(i).NumBlue
        tempColorDiff = Abs(intR1 - 212) + Abs(intG1 - 208) + Abs(intB1 - 200)
        If tempColorDiff < minColorDiff Then
            minColorDiff = tempColorDiff
            FindGray = i
        End If
    Next i
    If minColorDiff < 100 Then
        ColorPalette(FindGray).NumRed = 212
        ColorPalette(FindGray).NumGreen = 208
        ColorPalette(FindGray).NumBlue = 200
        ColorPalette(FindGray).NumOther = 0
        Exit Function
    End If
    If minColorDiff > 5 Then
        FindGray = 0
    End If
End Function

Public Sub WriteToLog(LogText As String)
    LOGGER.WriteLine (LogText)
End Sub

Public Sub ReadConfig()
    Dim begtext As Integer
    Dim endtext As Integer
    Dim Config As TextStream
    Dim ConfigFile As FileSystemObject
    
    On Error GoTo errHandler
    
    ConfigData.Language = "Default"
    ConfigData.MainLeft = 500
    ConfigData.MainTop = 500
    ConfigData.ImageOpen = "c:\"
    ConfigData.ImageSave = "c:\"
    ConfigData.Author = "Sticker Generator"
    
    If Not mcrMain.IsFileThere(App.Path & "\Settings\config.txt") Then
        MsgBox "Could not find: Config File " & App.Path & "\Settings\Config.txt." & Chr(10) & "Please be sure that all files are in the same folder."
        Call mcrMain.WriteToLog("Could not find Config File " & App.Path & "\Settings\Config.txt")
        Exit Sub
    End If
    
    Call mcrMain.WriteToLog("Attempting to read Config file.")
    
    Set ConfigFile = New FileSystemObject
    Set Config = LoggerFile.OpenTextFile(App.Path & "\Settings\config.txt", ForReading)

    ConfigData.ConfigString = Trim(Config.ReadAll)

    If ConfigData.ConfigString Like "*language*" Then
        begtext = InStr(ConfigData.ConfigString, "<language>") + 10
        endtext = InStrRev(ConfigData.ConfigString, "<language>")
        ConfigData.Language = Mid(ConfigData.ConfigString, begtext, endtext - begtext)
        
        begtext = InStr(ConfigData.ConfigString, "<left>") + 6
        endtext = InStrRev(ConfigData.ConfigString, "<left>")
        ConfigData.MainLeft = Mid(ConfigData.ConfigString, begtext, endtext - begtext)
        
        begtext = InStr(ConfigData.ConfigString, "<top>") + 5
        endtext = InStrRev(ConfigData.ConfigString, "<top>")
        ConfigData.MainTop = Mid(ConfigData.ConfigString, begtext, endtext - begtext)
                
        begtext = InStr(ConfigData.ConfigString, "<image>") + 7
        endtext = InStrRev(ConfigData.ConfigString, "<image>")
        ConfigData.ImageOpen = Mid(ConfigData.ConfigString, begtext, endtext - begtext)
                        
        begtext = InStr(ConfigData.ConfigString, "<save>") + 6
        endtext = InStrRev(ConfigData.ConfigString, "<save>")
        ConfigData.ImageSave = Mid(ConfigData.ConfigString, begtext, endtext - begtext)
                                
        begtext = InStr(ConfigData.ConfigString, "<author>") + 8
        endtext = InStrRev(ConfigData.ConfigString, "<author>")
        ConfigData.Author = Mid(ConfigData.ConfigString, begtext, endtext - begtext)
    Else
        Call mcrMain.WriteToLog("CONFIG FILE HAS BEEN CORRUPTED!")
        MsgBox "Config File has been corrupted.  A new config file will be written."
        Config.Close
        Set Config = Nothing
        Exit Sub
    End If
    
    Call mcrMain.WriteToLog("Config file successfully read.  Attempting to close file.")
    Config.Close
    Set Config = Nothing
    
    Exit Sub
    
errHandler:
    Screen.MousePointer = vbDefault
    MsgBox "ReadConfig has caused an error " & Err.Description
    Call WriteToLog("ERROR IN READCONFIG " & Err.Description)
    Config.Close
    Set Config = Nothing
End Sub

Public Sub WriteConfig()
    Dim Config As TextStream
    Dim ConfigFile As FileSystemObject
    
    On Error GoTo errHandler
    
    Call mcrMain.WriteToLog("Attempting to read Config file.")
    
    Set ConfigFile = New FileSystemObject
    Set Config = LoggerFile.OpenTextFile(App.Path & "\Settings\config.txt", ForWriting, True)

    ConfigData.ConfigString = "<language>" & ConfigData.Language & "<language>" & _
        "<left>" & ConfigData.MainLeft & "<left>" & _
        "<top>" & ConfigData.MainTop & "<top>" & _
        "<image>" & ConfigData.ImageOpen & "<image>" & _
        "<save>" & ConfigData.ImageSave & "<save>" & _
        "<author>" & ConfigData.Author & "<author>"
    
    Config.Write (ConfigData.ConfigString)
    
    Config.Close
    Set Config = Nothing
    
    Exit Sub
    
errHandler:
    Screen.MousePointer = vbDefault
    MsgBox "WriteConfig has caused an error " & Err.Description
    Call WriteToLog("ERROR IN WRITECONFIG " & Err.Description)
    Config.Close
    Set Config = Nothing
End Sub

Public Sub CreateLogger()
    Set LoggerFile = New FileSystemObject
    Set LOGGER = LoggerFile.OpenTextFile(App.Path & "\Settings\Log_File.txt", ForWriting, True)
    Call WriteToLog("Application Location " & App.Path)
End Sub

Public Function IsFileThere(FileName As String) As Boolean
    On Error Resume Next
    Dim i
    i = FreeFile
    Open FileName For Input As i
    If Err Then
        IsFileThere = False
    Else
        Close i
        IsFileThere = True
    End If
End Function
